@@ -1,7 +1,7 @@
 ////////////////////////////////////////////////////////////
 //
 // SFML - Simple and Fast Multimedia Library
-// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
+// Copyright (C) 2007-2008 Laurent Gomila (laurent.gom@gmail.com)
 //
 // This software is provided 'as-is', without any express or implied warranty.
 // In no event will the authors be held liable for any damages arising from the use of this software.
@@ -27,10 +27,8 @@
 // Headers
 ////////////////////////////////////////////////////////////
 #include <SFML/Graphics/PostFX.hpp>
-#include <SFML/Graphics/GraphicsDevice.hpp>
-#include <SFML/Graphics/OpenGL.hpp>
 #include <SFML/Graphics/RenderWindow.hpp>
-#include <SFML/Window/OpenGLCaps.hpp>
+#include <SFML/Graphics/Renderer.hpp>
 #include <fstream>
 #include <iostream>
 #include <set>
@@ -45,7 +43,8 @@ namespace sf
 PostFX::PostFX() :
 myShaderProgram(0)
 {
-
+    // No filtering on frame buffer
+    myFrameBuffer.SetSmooth(false);
 }
 
 
@@ -60,6 +59,9 @@ mySamplers      (Copy.mySamplers),
 myFragmentShader(Copy.myFragmentShader),
 myFrameBuffer   (Copy.myFrameBuffer)
 {
+    // No filtering on frame buffer
+    myFrameBuffer.SetSmooth(false);
+
     // Create the shaders and the program
     if (Copy.myShaderProgram)
         CreateProgram();
@@ -131,7 +133,7 @@ void PostFX::SetParameter(const std::string& Name, float X)
         if (Location != -1)
             GLCheck(glUniform1fARB(Location, X));
         else
-            std::cerr << "Parameter \"" << Name << "\" not found in effect" << std::endl;
+            std::cerr << "Parameter \"" << Name << "\" not found in post-effect" << std::endl;
 
         // Disable program
         GLCheck(glUseProgramObjectARB(0));
@@ -154,7 +156,7 @@ void PostFX::SetParameter(const std::string& Name, float X, float Y)
         if (Location != -1)
             GLCheck(glUniform2fARB(Location, X, Y));
         else
-            std::cerr << "Parameter \"" << Name << "\" not found in effect" << std::endl;
+            std::cerr << "Parameter \"" << Name << "\" not found in post-effect" << std::endl;
 
         // Disable program
         GLCheck(glUseProgramObjectARB(0));
@@ -177,7 +179,7 @@ void PostFX::SetParameter(const std::string& Name, float X, float Y, float Z)
         if (Location != -1)
             GLCheck(glUniform3fARB(Location, X, Y, Z));
         else
-            std::cerr << "Parameter \"" << Name << "\" not found in effect" << std::endl;
+            std::cerr << "Parameter \"" << Name << "\" not found in post-effect" << std::endl;
 
         // Disable program
         GLCheck(glUseProgramObjectARB(0));
@@ -200,7 +202,7 @@ void PostFX::SetParameter(const std::string& Name, float X, float Y, float Z, fl
         if (Location != -1)
             GLCheck(glUniform4fARB(Location, X, Y, Z, W));
         else
-            std::cerr << "Parameter \"" << Name << "\" not found in effect" << std::endl;
+            std::cerr << "Parameter \"" << Name << "\" not found in post-effect" << std::endl;
 
         // Disable program
         GLCheck(glUseProgramObjectARB(0));
@@ -239,10 +241,13 @@ PostFX& PostFX::operator =(const PostFX& Other)
 ////////////////////////////////////////////////////////////
 bool PostFX::CanUsePostFX()
 {
-    return OpenGLCaps::CheckExtension("GL_ARB_shading_language_100") &&
-           OpenGLCaps::CheckExtension("GL_ARB_shader_objects")       &&
-           OpenGLCaps::CheckExtension("GL_ARB_vertex_shader")        &&
-           OpenGLCaps::CheckExtension("GL_ARB_fragment_shader");
+    // Make sure the renderer exists
+    GetRenderer();
+
+    return glewIsSupported("GL_ARB_shading_language_100") != 0 &&
+           glewIsSupported("GL_ARB_shader_objects")       != 0 &&
+           glewIsSupported("GL_ARB_vertex_shader")        != 0 &&
+           glewIsSupported("GL_ARB_fragment_shader")      != 0;
 }
 
 
@@ -259,11 +264,9 @@ void PostFX::Render(const RenderWindow& Window) const
     if ((Window.GetWidth() != myFrameBuffer.GetWidth()) || (Window.GetHeight() != myFrameBuffer.GetHeight()))
     {
         myFrameBuffer.Resize(Window.GetWidth(), Window.GetHeight());
-        myFrameBuffer.SetSmooth(false);
-        myFrameBuffer.SetRepeat(false);
     }
 
-    // Copy the current framebuffer to our frame buffer texture
+    // Copy the current framebuffer pixels to our frame buffer texture
     myFrameBuffer.Bind();
     GLCheck(glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, myFrameBuffer.GetWidth(), myFrameBuffer.GetHeight()));
 
@@ -271,22 +274,23 @@ void PostFX::Render(const RenderWindow& Window) const
     GLCheck(glUseProgramObjectARB(myShaderProgram));
 
     // Bind textures
-    int Unit = 0;
+    GLint MaxUnits;
+    GLCheck(glGetIntegerv(GL_MAX_TEXTURE_COORDS_ARB, &MaxUnits));
+    unsigned int Unit = 0;
     for (std::map<std::string, Image*>::const_iterator i = mySamplers.begin(); i != mySamplers.end(); ++i)
     {
-        // Check that current texture unit is available
-        if (Unit >= OpenGLCaps::GetMaxTextureUnits())
+        // Check that the current texture unit is available
+        if (Unit >= static_cast<unsigned int>(MaxUnits))
         {
-            // Error : we have reached maximum texture units count
+            // Error : we have reached the maximum texture units count
             std::cerr << "Impossible to use texture \"" << i->first << "\" for post-effect : all available texture units are used" << std::endl;
             break;
         }
 
-        // Get the texture location in the effect
+        // Get texture location and assign it the given unit
         int Location = glGetUniformLocationARB(myShaderProgram, i->first.c_str());
         if (Location != -1)
         {
-            // Texture found : bind it and associate it to current texture unit
             GLCheck(glUniform1iARB(Location, Unit));
             GLCheck(glActiveTextureARB(GL_TEXTURE0_ARB + Unit));
             i->second->Bind();
@@ -294,29 +298,28 @@ void PostFX::Render(const RenderWindow& Window) const
         }
         else
         {
-            // Error : texture name not found in effect
-            std::cerr << "Texture \"" << i->first << "\" not found in effect" << std::endl;
+            std::cerr << "Texture \"" << i->first << "\" not found in post-effect" << std::endl;
         }
     }
 
     // Compute the texture coordinates (in case the texture is larger than the screen)
-    IntRect   FrameBufferRect(0, 0, myFrameBuffer.GetWidth(), myFrameBuffer.GetHeight());
+    IntRect FrameBufferRect(0, 0, myFrameBuffer.GetWidth(), myFrameBuffer.GetHeight());
     FloatRect TexCoords = myFrameBuffer.GetTexCoords(FrameBufferRect);
 
     // Render a fullscreen quad using the effect on our framebuffer
-    FloatRect Screen = Window.GetViewRect();
+    FloatRect Screen = Window.GetView().GetRect();
     glBegin(GL_QUADS);
-        glTexCoord2f(TexCoords.Left,  TexCoords.Top);    glVertex2f(Screen.Left,  Screen.Bottom);   
-        glTexCoord2f(TexCoords.Left,  TexCoords.Bottom); glVertex2f(Screen.Left,  Screen.Top);
-        glTexCoord2f(TexCoords.Right, TexCoords.Bottom); glVertex2f(Screen.Right, Screen.Top);
-        glTexCoord2f(TexCoords.Right, TexCoords.Top);    glVertex2f(Screen.Right, Screen.Bottom);   
+         glTexCoord2f(TexCoords.Left,  TexCoords.Top);    glVertex2f(Screen.Left,  Screen.Bottom);
+         glTexCoord2f(TexCoords.Right, TexCoords.Top);    glVertex2f(Screen.Right, Screen.Bottom);
+         glTexCoord2f(TexCoords.Right, TexCoords.Bottom); glVertex2f(Screen.Right, Screen.Top);
+         glTexCoord2f(TexCoords.Left,  TexCoords.Bottom); glVertex2f(Screen.Left,  Screen.Top);
     glEnd();
 
     // Disable program
     GLCheck(glUseProgramObjectARB(0));
 
     // Disable texture units
-    for (int i = 0; i < Unit; ++i)
+    for (unsigned int i = 0; i < Unit; ++i)
     {
         GLCheck(glActiveTextureARB(GL_TEXTURE0_ARB + i));
         GLCheck(glBindTexture(GL_TEXTURE_2D, 0));
@@ -404,46 +407,6 @@ std::string PostFX::PreprocessEffect(std::istream& File)
 }
 
 
-////////////////////////////////////////////////////////////
-/// Create a shader object from a file, and attach it to the program
-////////////////////////////////////////////////////////////
-bool PostFX::CreateAndAttachShader(const std::string& Source, unsigned int ShaderType)
-{
-    // Create shader object
-    GLhandleARB Shader = glCreateShaderObjectARB(ShaderType);
-
-    // Load source
-    int Length = static_cast<int>(Source.size());
-    const char* Data = Source.data();
-    GLCheck(glShaderSourceARB(Shader, 1, &Data, (const GLint*)&Length));
-
-    // Compile shader
-    GLCheck(glCompileShaderARB(Shader));
-
-    // Get compile log
-    int Success;
-    GLCheck(glGetObjectParameterivARB(Shader, GL_OBJECT_COMPILE_STATUS_ARB, (GLint*)&Success));
-    if (Success == GL_FALSE)
-    {
-        // Oops... compile errors !
-        char CompileLog[1024];
-        GLCheck(glGetInfoLogARB(Shader, sizeof(CompileLog), 0, CompileLog));
-        std::cerr << "Failed to compile effect :" << std::endl
-                  << CompileLog << std::endl;
-        GLCheck(glDeleteObjectARB(Shader));
-        return false;
-    }
-
-    // Attach the shader to the program
-    GLCheck(glAttachObjectARB(myShaderProgram, Shader));
-
-    // We can now delete our shader
-    GLCheck(glDeleteObjectARB(Shader));
-
-    return true;
-}
-
-
 ////////////////////////////////////////////////////////////
 /// Create the program and attach the shaders
 ////////////////////////////////////////////////////////////
@@ -452,7 +415,7 @@ void PostFX::CreateProgram()
     // Check that we can use post-FX !
     if (!CanUsePostFX())
     {
-        std::cerr << "Failed to create a sfPostFX : your system doesn't support effects" << std::endl;
+        std::cerr << "Failed to create a PostFX : your system doesn't support effects" << std::endl;
         return;
     }
 
@@ -460,48 +423,81 @@ void PostFX::CreateProgram()
     if (myShaderProgram)
         GLCheck(glDeleteObjectARB(myShaderProgram));
 
-    // Create the program
-    myShaderProgram = glCreateProgramObjectARB();
-
-    // Load vertex shader source (we provide it directly as it doesn't have to change)
-    static const std::string VertexShader =
+    // Define vertex shader source (we provide it directly as it doesn't have to change)
+    static const std::string VertexShaderSrc =
         "void main()"
         "{"
         "    gl_TexCoord[0] = gl_MultiTexCoord0;"
         "    gl_Position = ftransform();"
         "}";
 
-    // Create vertex shader
-    if (!CreateAndAttachShader(VertexShader, GL_VERTEX_SHADER_ARB))
+    // Create the program
+    myShaderProgram = glCreateProgramObjectARB();
+
+    // Create the shaders
+    GLhandleARB VertexShader   = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
+    GLhandleARB FragmentShader = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
+
+    // Compile them
+    const char* VertexSrc   = VertexShaderSrc.c_str();
+    const char* FragmentSrc = myFragmentShader.c_str();
+    GLCheck(glShaderSourceARB(VertexShader,   1, &VertexSrc,   NULL));
+    GLCheck(glShaderSourceARB(FragmentShader, 1, &FragmentSrc, NULL));
+    GLCheck(glCompileShaderARB(VertexShader));
+    GLCheck(glCompileShaderARB(FragmentShader));
+
+    // Check the compile logs
+    GLint Success;
+    GLCheck(glGetObjectParameterivARB(VertexShader, GL_OBJECT_COMPILE_STATUS_ARB, &Success));
+    if (Success == GL_FALSE)
     {
+        char CompileLog[1024];
+        GLCheck(glGetInfoLogARB(VertexShader, sizeof(CompileLog), 0, CompileLog));
+        std::cerr << "Failed to compile post-effect :" << std::endl
+                  << CompileLog << std::endl;
+        GLCheck(glDeleteObjectARB(VertexShader));
+        GLCheck(glDeleteObjectARB(FragmentShader));
+        GLCheck(glDeleteObjectARB(myShaderProgram));
+        myShaderProgram = 0;
+        return;
+    }
+    GLCheck(glGetObjectParameterivARB(FragmentShader, GL_OBJECT_COMPILE_STATUS_ARB, &Success));
+    if (Success == GL_FALSE)
+    {
+        char CompileLog[1024];
+        GLCheck(glGetInfoLogARB(FragmentShader, sizeof(CompileLog), 0, CompileLog));
+        std::cerr << "Failed to compile post-effect :" << std::endl
+                  << CompileLog << std::endl;
+        GLCheck(glDeleteObjectARB(VertexShader));
+        GLCheck(glDeleteObjectARB(FragmentShader));
         GLCheck(glDeleteObjectARB(myShaderProgram));
         myShaderProgram = 0;
         return;
     }
 
-    // Create fragment shader
-    if (!CreateAndAttachShader(myFragmentShader, GL_FRAGMENT_SHADER_ARB))
-    {
-        GLCheck(glDeleteObjectARB(myShaderProgram));
-        myShaderProgram = 0;
-        return;
-    }
+    // Attach the shaders to the program
+    GLCheck(glAttachObjectARB(myShaderProgram, VertexShader));
+    GLCheck(glAttachObjectARB(myShaderProgram, FragmentShader));
+
+    // We can now delete the shaders
+    GLCheck(glDeleteObjectARB(VertexShader));
+    GLCheck(glDeleteObjectARB(FragmentShader));
 
     // Link the program
     GLCheck(glLinkProgramARB(myShaderProgram));
 
     // Get link log
-    int Success;
-    GLCheck(glGetObjectParameterivARB(myShaderProgram, GL_OBJECT_LINK_STATUS_ARB, (GLint*)&Success));
+    GLCheck(glGetObjectParameterivARB(myShaderProgram, GL_OBJECT_LINK_STATUS_ARB, &Success));
     if (Success == GL_FALSE)
     {
-        // Oops... link errors !
+        // Oops... link errors
         char LinkLog[1024];
         GLCheck(glGetInfoLogARB(myShaderProgram, sizeof(LinkLog), 0, LinkLog));
-        std::cerr << "Failed to link effect :" << std::endl
+        std::cerr << "Failed to link post-effect :" << std::endl
                   << LinkLog << std::endl;
         GLCheck(glDeleteObjectARB(myShaderProgram));
         myShaderProgram = 0;
+        return;
     }
 }
 
