@@ -25,8 +25,8 @@
 ////////////////////////////////////////////////////////////
 // Headers
 ////////////////////////////////////////////////////////////
+#include <SFML/Window/WindowStyle.hpp> // important to be included first (conflict with None)
 #include <SFML/Window/Linux/WindowImplX11.hpp>
-#include <SFML/Window/WindowStyle.hpp>
 #include <SFML/Window/glext/glxext.h>
 #include <SFML/Window/glext/glext.h>
 #include <X11/keysym.h>
@@ -45,11 +45,12 @@ namespace priv
 ////////////////////////////////////////////////////////////
 Display*       WindowImplX11::ourDisplay          = NULL;
 int            WindowImplX11::ourScreen           = 0;
-WindowImplX11* WindowImplX11::ourDummyWindow      = NULL;
 WindowImplX11* WindowImplX11::ourFullscreenWindow = NULL;
 unsigned int   WindowImplX11::ourWindowsCount     = 0;
 unsigned long  WindowImplX11::ourEventMask        = FocusChangeMask | ButtonPressMask | ButtonReleaseMask | ButtonMotionMask |
                                                     PointerMotionMask | KeyPressMask | KeyReleaseMask | StructureNotifyMask;
+XIM            WindowImplX11::ourInputMethod      = NULL;
+
 
 ////////////////////////////////////////////////////////////
 /// Default constructor
@@ -61,7 +62,8 @@ myIsExternal  (false),
 myGLContext   (NULL),
 myAtomClose   (0),
 myOldVideoMode(-1),
-myHiddenCursor(0)
+myHiddenCursor(0),
+myInputContext(NULL)
 {
     // Open the display at first call
     if (!OpenDisplay())
@@ -73,7 +75,8 @@ myHiddenCursor(0)
 
     // Create the rendering context
     XVisualInfo Visual;
-    if (!CreateContext(VideoMode(myWidth, myHeight, 32), false, Visual, 0))
+    WindowSettings Params(0, 0, 0);
+    if (!CreateContext(VideoMode(myWidth, myHeight, 32), Visual, Params))
         return;
 
     // Create a new color map with the chosen visual
@@ -94,25 +97,22 @@ myHiddenCursor(0)
                              Visual.visual,
                              CWColormap, &Attributes);
 
-    // Save as the dummy window
-    if (!ourDummyWindow)
-        ourDummyWindow = this;
-
     // Set our context as the current OpenGL context for rendering
-    SetCurrent();
+    SetActive();
 }
 
 
 ////////////////////////////////////////////////////////////
 /// Create the window implementation from an existing control
 ////////////////////////////////////////////////////////////
-WindowImplX11::WindowImplX11(WindowHandle Handle, int AntialiasingLevel) :
+WindowImplX11::WindowImplX11(WindowHandle Handle, WindowSettings& Params) :
 myWindow      (0),
 myIsExternal  (true),
 myGLContext   (NULL),
 myAtomClose   (0),
 myOldVideoMode(-1),
-myHiddenCursor(0)
+myHiddenCursor(0),
+myInputContext(NULL)
 {
     // Open the display at first call
     if (!OpenDisplay())
@@ -140,11 +140,9 @@ myHiddenCursor(0)
         unsigned long Mask = VisualDepthMask | VisualIDMask;
 
         // Create the rendering context
-        VideoMode Mode = VideoMode::GetDesktopMode();
-        Mode.Width  = myWidth;
-        Mode.Height = myHeight;
+        VideoMode Mode(myWidth, myHeight, VideoMode::GetDesktopMode().BitsPerPixel);
         XVisualInfo Visual;
-        if (!CreateContext(Mode, false, Visual, AntialiasingLevel, Template, Mask))
+        if (!CreateContext(Mode, Visual, Params, Template, Mask))
             return;
 
         // Create a new color map with the chosen visual
@@ -163,13 +161,14 @@ myHiddenCursor(0)
 ////////////////////////////////////////////////////////////
 /// Create the window implementation
 ////////////////////////////////////////////////////////////
-WindowImplX11::WindowImplX11(VideoMode Mode, const std::string& Title, unsigned long WindowStyle, int AntialiasingLevel) :
+WindowImplX11::WindowImplX11(VideoMode Mode, const std::string& Title, unsigned long WindowStyle, WindowSettings& Params) :
 myWindow      (0),
 myIsExternal  (false),
 myGLContext   (NULL),
 myAtomClose   (0),
 myOldVideoMode(-1),
-myHiddenCursor(0)
+myHiddenCursor(0),
+myInputContext(NULL)
 {
     // Open the display at first call
     if (!OpenDisplay())
@@ -191,9 +190,13 @@ myHiddenCursor(0)
     int Width  = myWidth  = Mode.Width;
     int Height = myHeight = Mode.Height;
 
+    // Switch to fullscreen if necessary
+    if (Fullscreen)
+        SwitchToFullscreen(Mode);
+
     // Create the rendering context
     XVisualInfo Visual;
-    if (!CreateContext(Mode, Fullscreen, Visual, AntialiasingLevel))
+    if (!CreateContext(Mode, Visual, Params))
         return;
 
     // Create a new color map with the chosen visual
@@ -259,9 +262,14 @@ myHiddenCursor(0)
     
             WMHints Hints;
             Hints.Flags       = MWM_HINTS_FUNCTIONS | MWM_HINTS_DECORATIONS;
-            Hints.Decorations = MWM_DECOR_BORDER | MWM_DECOR_TITLE | MWM_DECOR_MINIMIZE | MWM_DECOR_MENU;
-            Hints.Functions   = MWM_FUNC_MOVE | MWM_FUNC_MINIMIZE;
-    
+            Hints.Decorations = 0;
+            Hints.Functions   = 0;
+
+            if (WindowStyle & Style::Titlebar)
+            {
+                Hints.Decorations |= MWM_DECOR_BORDER | MWM_DECOR_TITLE | MWM_DECOR_MINIMIZE | MWM_DECOR_MENU;
+                Hints.Functions   |= MWM_FUNC_MOVE | MWM_FUNC_MINIMIZE;
+            }
             if (WindowStyle & Style::Resize)
             {
                 Hints.Decorations |= MWM_DECOR_MAXIMIZE | MWM_DECOR_RESIZEH;
@@ -308,9 +316,18 @@ WindowImplX11::~WindowImplX11()
     // Cleanup graphical resources
     CleanUp();
 
+    // Destroy the input context
+    if (myInputContext)
+    {
+        XDestroyIC(myInputContext);
+    }
+
     // Destroy the window
     if (myWindow && !myIsExternal)
+    {
         XDestroyWindow(ourDisplay, myWindow);
+        XFlush(ourDisplay);
+    }
 
     // Close the display
     CloseDisplay();
@@ -340,12 +357,19 @@ void WindowImplX11::ProcessEvents()
 
 
 ////////////////////////////////////////////////////////////
-/// /see WindowImpl::MakeCurrent
+/// /see WindowImpl::MakeActive
 ////////////////////////////////////////////////////////////
-void WindowImplX11::MakeCurrent() const
+void WindowImplX11::MakeActive(bool Active) const
 {
-    if (myWindow && myGLContext)
-        glXMakeCurrent(ourDisplay, myWindow, myGLContext);
+    if (Active)
+    {
+        if (myWindow && myGLContext)
+            glXMakeCurrent(ourDisplay, myWindow, myGLContext);
+    }
+    else
+    {
+        glXMakeCurrent(ourDisplay, None, NULL);
+    }
 }
 
 
@@ -367,6 +391,7 @@ void WindowImplX11::UseVerticalSync(bool Enabled)
 void WindowImplX11::ShowMouseCursor(bool Show)
 {
     XDefineCursor(ourDisplay, myWindow, Show ? None : myHiddenCursor);
+    XFlush(ourDisplay);
 }
 
 
@@ -376,6 +401,7 @@ void WindowImplX11::ShowMouseCursor(bool Show)
 void WindowImplX11::SetCursorPosition(unsigned int Left, unsigned int Top)
 {
     XWarpPointer(ourDisplay, None, myWindow, 0, 0, 0, 0, Left, Top);
+    XFlush(ourDisplay);
 }
 
 
@@ -385,6 +411,7 @@ void WindowImplX11::SetCursorPosition(unsigned int Left, unsigned int Top)
 void WindowImplX11::SetPosition(int Left, int Top)
 {
     XMoveWindow(ourDisplay, myWindow, Left, Top);
+    XFlush(ourDisplay);
 }
 
 
@@ -403,61 +430,78 @@ void WindowImplX11::Show(bool State)
 
 
 ////////////////////////////////////////////////////////////
-/// Create the OpenGL rendering context
+/// /see sfWindowImpl::EnableKeyRepeat
 ////////////////////////////////////////////////////////////
-bool WindowImplX11::CreateContext(VideoMode Mode, bool Fullscreen, XVisualInfo& ChosenVisual, int AntialiasingLevel, XVisualInfo Template, unsigned long Mask)
+void WindowImplX11::EnableKeyRepeat(bool Enabled)
 {
-    // Fullscreen mode
-    if (Fullscreen)
+    if (Enabled)
+        XAutoRepeatOn(ourDisplay);
+    else
+        XAutoRepeatOff(ourDisplay);
+
+    XFlush(ourDisplay);
+}
+
+
+////////////////////////////////////////////////////////////
+/// Switch to fullscreen mode
+////////////////////////////////////////////////////////////
+void WindowImplX11::SwitchToFullscreen(const VideoMode& Mode)
+{
+    // Check if the XRandR extension is present
+    int Version;
+    if (XQueryExtension(ourDisplay, "RANDR", &Version, &Version, &Version))
     {
-        // Check if the XRandR extension is present
-        int Version;
-        if (XQueryExtension(ourDisplay, "RANDR", &Version, &Version, &Version))
+        // Get the current configuration
+        XRRScreenConfiguration* Config = XRRGetScreenInfo(ourDisplay, RootWindow(ourDisplay, ourScreen));
+        if (Config)
         {
-            // Get the current configuration
-            XRRScreenConfiguration* Config = XRRGetScreenInfo(ourDisplay, RootWindow(ourDisplay, ourScreen));
-            if (Config)
+            // Get the current rotation
+            Rotation CurrentRotation;
+            myOldVideoMode = XRRConfigCurrentConfiguration(Config, &CurrentRotation);
+
+            // Get the available screen sizes
+            int NbSizes;
+            XRRScreenSize* Sizes = XRRConfigSizes(Config, &NbSizes);
+            if (Sizes && (NbSizes > 0))
             {
-                // Get the current rotation
-                Rotation CurrentRotation;
-                myOldVideoMode = XRRConfigCurrentConfiguration(Config, &CurrentRotation);
-
-                // Get the available screen sizes
-                int NbSizes;
-                XRRScreenSize* Sizes = XRRConfigSizes(Config, &NbSizes);
-                if (Sizes && (NbSizes > 0))
+                // Search a matching size
+                for (int i = 0; i < NbSizes; ++i)
                 {
-                    // Search a matching size
-                    for (int i = 0; i < NbSizes; ++i)
+                    if ((Sizes[i].width == static_cast<int>(Mode.Width)) && (Sizes[i].height == static_cast<int>(Mode.Height)))
                     {
-                        if ((Sizes[i].width == static_cast<int>(Mode.Width)) && (Sizes[i].height == static_cast<int>(Mode.Height)))
-                        {
-                            // Switch to fullscreen mode
-                            XRRSetScreenConfig(ourDisplay, Config, RootWindow(ourDisplay, ourScreen), i, CurrentRotation, CurrentTime);
+                        // Switch to fullscreen mode
+                        XRRSetScreenConfig(ourDisplay, Config, RootWindow(ourDisplay, ourScreen), i, CurrentRotation, CurrentTime);
 
-                            // Set "this" as the current fullscreen window
-                            ourFullscreenWindow = this;
-                            break;
-                        }
+                        // Set "this" as the current fullscreen window
+                        ourFullscreenWindow = this;
+                        break;
                     }
                 }
+            }
 
-                // Free the configuration instance
-                XRRFreeScreenConfigInfo(Config);
-            }
-            else
-            {
-                // Failed to get the screen configuration
-                std::cerr << "Failed to get the current screen configuration for fullscreen mode, switching to windiw mode" << std::endl;
-            }
+            // Free the configuration instance
+            XRRFreeScreenConfigInfo(Config);
         }
         else
         {
-            // XRandr extension is not supported : we cannot use fullscreen mode
-            std::cerr << "Fullscreen is not supported, switching to window mode" << std::endl;
+            // Failed to get the screen configuration
+            std::cerr << "Failed to get the current screen configuration for fullscreen mode, switching to windiw mode" << std::endl;
         }
     }
+    else
+    {
+        // XRandr extension is not supported : we cannot use fullscreen mode
+        std::cerr << "Fullscreen is not supported, switching to window mode" << std::endl;
+    }
+}
 
+
+////////////////////////////////////////////////////////////
+/// Create the OpenGL rendering context
+////////////////////////////////////////////////////////////
+bool WindowImplX11::CreateContext(const VideoMode& Mode, XVisualInfo& ChosenVisual, WindowSettings& Params, XVisualInfo Template, unsigned long Mask)
+{
     // Get all the visuals matching the template
     Template.screen = ourScreen;
     int NbVisuals = 0;
@@ -470,61 +514,56 @@ bool WindowImplX11::CreateContext(VideoMode Mode, bool Fullscreen, XVisualInfo&
         return false;
     }
 
-    // A helper structure for storing visual attributes
-    struct VisualAttributes
-    {
-        int RGBA, DoubleBuffer, Red, Green, Blue, Alpha, Depth, Stencil, MultiSampling, Samples;
-    };
-
     // Find the best visual
-    bool Found = false;
-    XVisualInfo* Visual = &Visuals[0];
-    while (!Found)
+    int          BestScore  = 0xFFFF;
+    XVisualInfo* BestVisual = NULL;
+    while (!BestVisual)
     {
-        VisualAttributes BestAttribs = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
-        int MultiSampling = (AntialiasingLevel > 0 ? 1 : 0);
-
         for (int i = 0; i < NbVisuals; ++i)
         {
             // Get the current visual attributes
-            VisualAttributes Attribs = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
-            glXGetConfig(ourDisplay, &Visuals[i], GLX_RGBA,               &Attribs.RGBA);
-            glXGetConfig(ourDisplay, &Visuals[i], GLX_DOUBLEBUFFER,       &Attribs.DoubleBuffer); 
-            glXGetConfig(ourDisplay, &Visuals[i], GLX_RED_SIZE,           &Attribs.Red);
-            glXGetConfig(ourDisplay, &Visuals[i], GLX_GREEN_SIZE,         &Attribs.Green); 
-            glXGetConfig(ourDisplay, &Visuals[i], GLX_BLUE_SIZE,          &Attribs.Blue); 
-            glXGetConfig(ourDisplay, &Visuals[i], GLX_ALPHA_SIZE,         &Attribs.Alpha); 
-            glXGetConfig(ourDisplay, &Visuals[i], GLX_DEPTH_SIZE,         &Attribs.Depth);        
-            glXGetConfig(ourDisplay, &Visuals[i], GLX_STENCIL_SIZE,       &Attribs.Stencil);
-            glXGetConfig(ourDisplay, &Visuals[i], GLX_SAMPLE_BUFFERS_ARB, &Attribs.MultiSampling);        
-            glXGetConfig(ourDisplay, &Visuals[i], GLX_SAMPLES_ARB,        &Attribs.Samples);
-    
-            // Compare with the current best
-            if (((AntialiasingLevel == 0) || (Attribs.MultiSampling == MultiSampling) && (Attribs.Samples == AntialiasingLevel)) &&
-               ((BestAttribs.RGBA    < Attribs.RGBA)    || (BestAttribs.DoubleBuffer < Attribs.DoubleBuffer) ||
-                (BestAttribs.Red     < Attribs.Red)     || (BestAttribs.Green        < Attribs.Green)        ||
-                (BestAttribs.Blue    < Attribs.Blue)    || (BestAttribs.Alpha        < Attribs.Alpha)        ||
-                (BestAttribs.Stencil < Attribs.Stencil) || (BestAttribs.Depth        < Attribs.Depth)))
+            int RGBA, DoubleBuffer, Red, Green, Blue, Alpha, Depth, Stencil, MultiSampling, Samples;
+            glXGetConfig(ourDisplay, &Visuals[i], GLX_RGBA,               &RGBA);
+            glXGetConfig(ourDisplay, &Visuals[i], GLX_DOUBLEBUFFER,       &DoubleBuffer); 
+            glXGetConfig(ourDisplay, &Visuals[i], GLX_RED_SIZE,           &Red);
+            glXGetConfig(ourDisplay, &Visuals[i], GLX_GREEN_SIZE,         &Green); 
+            glXGetConfig(ourDisplay, &Visuals[i], GLX_BLUE_SIZE,          &Blue); 
+            glXGetConfig(ourDisplay, &Visuals[i], GLX_ALPHA_SIZE,         &Alpha); 
+            glXGetConfig(ourDisplay, &Visuals[i], GLX_DEPTH_SIZE,         &Depth);        
+            glXGetConfig(ourDisplay, &Visuals[i], GLX_STENCIL_SIZE,       &Stencil);
+            glXGetConfig(ourDisplay, &Visuals[i], GLX_SAMPLE_BUFFERS_ARB, &MultiSampling);        
+            glXGetConfig(ourDisplay, &Visuals[i], GLX_SAMPLES_ARB,        &Samples);
+
+            // First check the mandatory parameters
+            if ((RGBA == 0) || (DoubleBuffer == 0))
+                continue;
+
+            // Evaluate the current configuration
+            int Color = Red + Green + Blue + Alpha;
+            int Score = EvaluateConfig(Mode, Params, Color, Depth, Stencil, MultiSampling ? Samples : 0);
+
+            // Keep it if it's better than the current best
+            if (Score < BestScore)
             {
-                BestAttribs = Attribs;
-                Visual      = &Visuals[i];
-                Found       = true;
+                BestScore  = Score;
+                BestVisual = &Visuals[i];
+                break;
             }
         }
 
         // If no visual has been found, try a lower level of antialiasing
-        if (!Found)
+        if (!BestVisual)
         {
-            if (AntialiasingLevel > 2)
+            if (Params.AntialiasingLevel > 2)
             {
                 std::cerr << "Failed to find a pixel format supporting "
-                          << AntialiasingLevel << " antialiasing levels ; trying with 2 levels" << std::endl;
-                AntialiasingLevel = 2;
+                          << Params.AntialiasingLevel << " antialiasing levels ; trying with 2 levels" << std::endl;
+                Params.AntialiasingLevel = 2;
             }
-            else if (AntialiasingLevel > 0)
+            else if (Params.AntialiasingLevel > 0)
             {
                 std::cerr << "Failed to find a pixel format supporting antialiasing ; antialiasing will be disabled" << std::endl;
-                AntialiasingLevel = 0;
+                Params.AntialiasingLevel = 0;
             }
             else
             {
@@ -535,26 +574,26 @@ bool WindowImplX11::CreateContext(VideoMode Mode, bool Fullscreen, XVisualInfo&
     }
 
     // Create the OpenGL context
-    myGLContext = glXCreateContext(ourDisplay, Visual, glXGetCurrentContext(), true);
+    myGLContext = glXCreateContext(ourDisplay, BestVisual, glXGetCurrentContext(), true);
     if (myGLContext == NULL)
     {
         std::cerr << "Failed to create an OpenGL context for this window" << std::endl;
         return false;
     }
 
-    // Extract the depth and stencil bits from the chosen format
+    // Update the creation settings from the chosen format
     int Depth, Stencil;
-    glXGetConfig(ourDisplay, Visual, GLX_DEPTH_SIZE,   &Depth);
-    glXGetConfig(ourDisplay, Visual, GLX_STENCIL_SIZE, &Stencil);
-    myDepthBits   = static_cast<unsigned int>(Depth);
-    myStencilBits = static_cast<unsigned int>(Stencil);
+    glXGetConfig(ourDisplay, BestVisual, GLX_DEPTH_SIZE,   &Depth);
+    glXGetConfig(ourDisplay, BestVisual, GLX_STENCIL_SIZE, &Stencil);
+    Params.DepthBits   = static_cast<unsigned int>(Depth);
+    Params.StencilBits = static_cast<unsigned int>(Stencil);
 
     // Assign the chosen visual, and free the temporary visuals array
-    ChosenVisual = *Visual;
+    ChosenVisual = *BestVisual;
     XFree(Visuals);
-    
+
     // Enable multisampling if needed
-    if (AntialiasingLevel > 0)
+    if (Params.AntialiasingLevel > 0)
         glEnable(GL_MULTISAMPLE_ARB);
 
     return true;
@@ -570,6 +609,21 @@ void WindowImplX11::Initialize()
     myAtomClose = XInternAtom(ourDisplay, "WM_DELETE_WINDOW", false);
     XSetWMProtocols(ourDisplay, myWindow, &myAtomClose, 1);
 
+    // Create the input context
+    if (ourInputMethod)
+    {
+        myInputContext = XCreateIC(ourInputMethod,
+                                   XNClientWindow,  myWindow,
+                                   XNFocusWindow,   myWindow,
+                                   XNInputStyle,    XIMPreeditNothing  | XIMStatusNothing,
+                                   //XNResourceName,  "SFML",
+                                   //XNResourceClass, "SFML",
+                                   NULL);
+        
+        if (!myInputContext)
+            std::cerr << "Failed to create input context for window -- TextEntered event won't be able to return unicode" << std::endl;
+    }
+
     // Show the window
     XMapWindow(ourDisplay, myWindow);
     XFlush(ourDisplay);
@@ -578,7 +632,7 @@ void WindowImplX11::Initialize()
     CreateHiddenCursor();
 
     // Set our context as the current OpenGL context for rendering
-    SetCurrent();
+    SetActive();
 
     // Flush the commands queue
     XFlush(ourDisplay);
@@ -640,19 +694,6 @@ void WindowImplX11::CleanUp()
     // Destroy the OpenGL context
     if (myGLContext)
     {
-        // If this is not the dummy window, we must set it as the valid rendering context to avoid a crash with next OpenGL command
-        if (this != ourDummyWindow)
-        {
-            if (ourDummyWindow)
-                ourDummyWindow->SetCurrent();
-        }
-        else
-        {
-            ourDummyWindow = NULL;
-            glXMakeCurrent(ourDisplay, 0, NULL);
-        }
-
-        // Now that our context is no more the current one, we can destroy it
         glXDestroyContext(ourDisplay, myGLContext);
         myGLContext = NULL;
     }
@@ -666,7 +707,7 @@ void WindowImplX11::CleanUp()
 Bool WindowImplX11::CheckEvent(::Display*, XEvent* Event, XPointer UserData)
 {
     // Just check if the event matches our window
-    return Event->xany.window == reinterpret_cast<Window>(UserData);
+    return Event->xany.window == reinterpret_cast< ::Window >(UserData);
 }
 
 
@@ -688,6 +729,10 @@ void WindowImplX11::ProcessEvent(XEvent WinEvent)
         // Gain focus event
         case FocusIn :
         {
+            // Update the input context
+            if (myInputContext)
+                XSetICFocus(myInputContext);
+
             Event Evt;
             Evt.Type = Event::GainedFocus;
             SendEvent(Evt);
@@ -697,6 +742,10 @@ void WindowImplX11::ProcessEvent(XEvent WinEvent)
         // Lost focus event
         case FocusOut :
         {
+            // Update the input context
+            if (myInputContext)
+                XUnsetICFocus(myInputContext);
+
             Event Evt;
             Evt.Type = Event::LostFocus;
             SendEvent(Evt);
@@ -728,7 +777,6 @@ void WindowImplX11::ProcessEvent(XEvent WinEvent)
                 Event Evt;
                 Evt.Type = Event::Closed;
                 SendEvent(Evt);
-                CleanUp();
             }
             break;
         }
@@ -737,9 +785,10 @@ void WindowImplX11::ProcessEvent(XEvent WinEvent)
         case KeyPress :
         {
             // Get the keysym of the key that has been pressed
+            static XComposeStatus KeyboardStatus;
             char Buffer[32];
             KeySym Sym;
-            XLookupString(&WinEvent.xkey, Buffer, 32, &Sym, NULL);
+            XLookupString(&WinEvent.xkey, Buffer, sizeof(Buffer), &Sym, &KeyboardStatus);
 
             // Fill the event parameters
             Event Evt;
@@ -748,10 +797,39 @@ void WindowImplX11::ProcessEvent(XEvent WinEvent)
             Evt.Key.Alt     = WinEvent.xkey.state & Mod1Mask;
             Evt.Key.Control = WinEvent.xkey.state & ControlMask;
             Evt.Key.Shift   = WinEvent.xkey.state & ShiftMask;
-
-            // TODO : Generate a TextEntered event
-
             SendEvent(Evt);
+
+            // Generate a TextEntered event
+            if (!XFilterEvent(&WinEvent, None))
+            {
+                #ifdef X_HAVE_UTF8_STRING
+                if (myInputContext)
+                {
+                    Status ReturnedStatus;
+                    Uint8  KeyBuffer[16];
+                    if (Xutf8LookupString(myInputContext, &WinEvent.xkey, reinterpret_cast<char*>(KeyBuffer), sizeof(KeyBuffer), NULL, &ReturnedStatus))
+                    {
+                        Event TextEvt;
+                        TextEvt.Type         = Event::TextEntered;
+                        TextEvt.Text.Unicode = UTF8toUTF16(KeyBuffer);
+                        SendEvent(TextEvt);
+                    }
+                }
+                else
+                #endif
+                {
+                    static XComposeStatus ComposeStatus;
+                    char KeyBuffer[16];
+                    if (XLookupString(&WinEvent.xkey, KeyBuffer, sizeof(KeyBuffer), NULL, &ComposeStatus))
+                    {
+                        Event TextEvt;
+                        TextEvt.Type         = Event::TextEntered;
+                        TextEvt.Text.Unicode = KeyBuffer[0];
+                        SendEvent(TextEvt);
+                    }
+                }
+            }
+
             break;
         }
 
@@ -849,94 +927,108 @@ Key::Code WindowImplX11::KeysymToSF(KeySym Sym)
 
     switch (Key)
     {
-        case XK_Escape :      return Key::Escape;
-        case XK_Shift_L :     return Key::Shift;
-        case XK_Shift_R :     return Key::Shift;
-        case XK_Control_L :   return Key::Control;
-        case XK_Control_R :   return Key::Control;
-        case XK_Alt_L :       return Key::Alt;
-        case XK_Alt_R :       return Key::Alt;
-        case XK_space :       return Key::Space;
-        case XK_Return :      return Key::Return;
-        case XK_KP_Enter :    return Key::Return;
-        case XK_BackSpace :   return Key::Back;
-        case XK_Tab :         return Key::Tab;
-        case XK_Prior :       return Key::PageUp;
-        case XK_Next :        return Key::PageDown;
-        case XK_End :         return Key::End;
-        case XK_Home :        return Key::Home;
-        case XK_Insert :      return Key::Insert;
-        case XK_Delete :      return Key::Delete;
-        case XK_KP_Add :      return Key::Add;
-        case XK_KP_Subtract : return Key::Subtract;
-        case XK_KP_Multiply : return Key::Multiply;
-        case XK_KP_Divide :   return Key::Divide;
-        case XK_Pause :       return Key::Pause;
-        case XK_F1 :          return Key::F1;
-        case XK_F2 :          return Key::F2;
-        case XK_F3 :          return Key::F3;
-        case XK_F4 :          return Key::F4;
-        case XK_F5 :          return Key::F5;
-        case XK_F6 :          return Key::F6;
-        case XK_F7 :          return Key::F7;
-        case XK_F8 :          return Key::F8;
-        case XK_F9 :          return Key::F9;
-        case XK_F10 :         return Key::F10;
-        case XK_F11 :         return Key::F11;
-        case XK_F12 :         return Key::F12;
-        case XK_F13 :         return Key::F13;
-        case XK_F14 :         return Key::F14;
-        case XK_F15 :         return Key::F15;
-        case XK_Left :        return Key::Left;
-        case XK_Right :       return Key::Right;
-        case XK_Up :          return Key::Up;
-        case XK_Down :        return Key::Down;
-        case XK_KP_0 :        return Key::Numpad0;
-        case XK_KP_1 :        return Key::Numpad1;
-        case XK_KP_2 :        return Key::Numpad2;
-        case XK_KP_3 :        return Key::Numpad3;
-        case XK_KP_4 :        return Key::Numpad4;
-        case XK_KP_5 :        return Key::Numpad5;
-        case XK_KP_6 :        return Key::Numpad6;
-        case XK_KP_7 :        return Key::Numpad7;
-        case XK_KP_8 :        return Key::Numpad8;
-        case XK_KP_9 :        return Key::Numpad9;
-        case XK_A :           return Key::A;
-        case XK_Z :           return Key::Z;
-        case XK_E :           return Key::E;
-        case XK_R :           return Key::R;
-        case XK_T :           return Key::T;
-        case XK_Y :           return Key::Y;
-        case XK_U :           return Key::U;
-        case XK_I :           return Key::I;
-        case XK_O :           return Key::O;
-        case XK_P :           return Key::P;
-        case XK_Q :           return Key::Q;
-        case XK_S :           return Key::S;
-        case XK_D :           return Key::D;
-        case XK_F :           return Key::F;
-        case XK_G :           return Key::G;
-        case XK_H :           return Key::H;
-        case XK_J :           return Key::J;
-        case XK_K :           return Key::K;
-        case XK_L :           return Key::L;
-        case XK_M :           return Key::M;
-        case XK_W :           return Key::W;
-        case XK_X :           return Key::X;
-        case XK_C :           return Key::C;
-        case XK_V :           return Key::V;
-        case XK_B :           return Key::B;
-        case XK_N :           return Key::N;
-        case XK_0 :           return Key::Num0;
-        case XK_1 :           return Key::Num1;
-        case XK_2 :           return Key::Num2;
-        case XK_3 :           return Key::Num3;
-        case XK_4 :           return Key::Num4;
-        case XK_5 :           return Key::Num5;
-        case XK_6 :           return Key::Num6;
-        case XK_7 :           return Key::Num7;
-        case XK_8 :           return Key::Num8;
-        case XK_9 :           return Key::Num9;
+        case XK_Shift_L :      return Key::LShift;
+        case XK_Shift_R :      return Key::RShift;
+        case XK_Control_L :    return Key::LControl;
+        case XK_Control_R :    return Key::RControl;
+        case XK_Alt_L :        return Key::LAlt;
+        case XK_Alt_R :        return Key::RAlt;
+        case XK_Super_L :      return Key::LSystem;
+        case XK_Super_R :      return Key::RSystem;
+        case XK_Menu :         return Key::Menu;
+        case XK_Escape :       return Key::Escape;
+        case XK_semicolon :    return Key::SemiColon;
+        case XK_slash :        return Key::Slash;
+        case XK_equal :        return Key::Equal;
+        case XK_minus :        return Key::Dash;
+        case XK_bracketleft :  return Key::LBracket;
+        case XK_bracketright : return Key::RBracket;
+        case XK_comma :        return Key::Comma;
+        case XK_period :       return Key::Period;
+        case XK_dead_acute :   return Key::Quote;
+        case XK_backslash :    return Key::BackSlash;
+        case XK_dead_grave :   return Key::Tilde;
+        case XK_space :        return Key::Space;
+        case XK_Return :       return Key::Return;
+        case XK_KP_Enter :     return Key::Return;
+        case XK_BackSpace :    return Key::Back;
+        case XK_Tab :          return Key::Tab;
+        case XK_Prior :        return Key::PageUp;
+        case XK_Next :         return Key::PageDown;
+        case XK_End :          return Key::End;
+        case XK_Home :         return Key::Home;
+        case XK_Insert :       return Key::Insert;
+        case XK_Delete :       return Key::Delete;
+        case XK_KP_Add :       return Key::Add;
+        case XK_KP_Subtract :  return Key::Subtract;
+        case XK_KP_Multiply :  return Key::Multiply;
+        case XK_KP_Divide :    return Key::Divide;
+        case XK_Pause :        return Key::Pause;
+        case XK_F1 :           return Key::F1;
+        case XK_F2 :           return Key::F2;
+        case XK_F3 :           return Key::F3;
+        case XK_F4 :           return Key::F4;
+        case XK_F5 :           return Key::F5;
+        case XK_F6 :           return Key::F6;
+        case XK_F7 :           return Key::F7;
+        case XK_F8 :           return Key::F8;
+        case XK_F9 :           return Key::F9;
+        case XK_F10 :          return Key::F10;
+        case XK_F11 :          return Key::F11;
+        case XK_F12 :          return Key::F12;
+        case XK_F13 :          return Key::F13;
+        case XK_F14 :          return Key::F14;
+        case XK_F15 :          return Key::F15;
+        case XK_Left :         return Key::Left;
+        case XK_Right :        return Key::Right;
+        case XK_Up :           return Key::Up;
+        case XK_Down :         return Key::Down;
+        case XK_KP_0 :         return Key::Numpad0;
+        case XK_KP_1 :         return Key::Numpad1;
+        case XK_KP_2 :         return Key::Numpad2;
+        case XK_KP_3 :         return Key::Numpad3;
+        case XK_KP_4 :         return Key::Numpad4;
+        case XK_KP_5 :         return Key::Numpad5;
+        case XK_KP_6 :         return Key::Numpad6;
+        case XK_KP_7 :         return Key::Numpad7;
+        case XK_KP_8 :         return Key::Numpad8;
+        case XK_KP_9 :         return Key::Numpad9;
+        case XK_A :            return Key::A;
+        case XK_Z :            return Key::Z;
+        case XK_E :            return Key::E;
+        case XK_R :            return Key::R;
+        case XK_T :            return Key::T;
+        case XK_Y :            return Key::Y;
+        case XK_U :            return Key::U;
+        case XK_I :            return Key::I;
+        case XK_O :            return Key::O;
+        case XK_P :            return Key::P;
+        case XK_Q :            return Key::Q;
+        case XK_S :            return Key::S;
+        case XK_D :            return Key::D;
+        case XK_F :            return Key::F;
+        case XK_G :            return Key::G;
+        case XK_H :            return Key::H;
+        case XK_J :            return Key::J;
+        case XK_K :            return Key::K;
+        case XK_L :            return Key::L;
+        case XK_M :            return Key::M;
+        case XK_W :            return Key::W;
+        case XK_X :            return Key::X;
+        case XK_C :            return Key::C;
+        case XK_V :            return Key::V;
+        case XK_B :            return Key::B;
+        case XK_N :            return Key::N;
+        case XK_0 :            return Key::Num0;
+        case XK_1 :            return Key::Num1;
+        case XK_2 :            return Key::Num2;
+        case XK_3 :            return Key::Num3;
+        case XK_4 :            return Key::Num4;
+        case XK_5 :            return Key::Num5;
+        case XK_6 :            return Key::Num6;
+        case XK_7 :            return Key::Num7;
+        case XK_8 :            return Key::Num8;
+        case XK_9 :            return Key::Num9;
     }
 
     return Key::Code(0);
@@ -956,10 +1048,13 @@ bool WindowImplX11::OpenDisplay(bool AddWindow)
         {
             ourScreen = DefaultScreen(ourDisplay);
 
-            // Setup the keyboard auto-repeat to behave like in Windows (only repeat KeyPress)
+            // Disable repetition of KeyRelease events when a key is held down
             Bool Supported;
             XkbSetDetectableAutoRepeat(ourDisplay, True, &Supported);
-            XFlush(ourDisplay);  
+            XFlush(ourDisplay);
+
+            // Get the input method (XIM) object
+            ourInputMethod = XOpenIM(ourDisplay, NULL, NULL, NULL);
         }
         else
         {
@@ -986,6 +1081,10 @@ void WindowImplX11::CloseDisplay()
     // If all windows have been destroyed, then we can close the display
     if (ourWindowsCount == 0)
     {
+        // Close the input method object
+        if (ourInputMethod)
+            XCloseIM(ourInputMethod);
+
         XCloseDisplay(ourDisplay);
         ourDisplay = NULL;
     }
