@@ -1,7 +1,7 @@
 ////////////////////////////////////////////////////////////
 //
 // SFML - Simple and Fast Multimedia Library
-// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
+// Copyright (C) 2007-2008 Laurent Gomila (laurent.gom@gmail.com)
 //
 // This software is provided 'as-is', without any express or implied warranty.
 // In no event will the authors be held liable for any damages arising from the use of this software.
@@ -28,6 +28,7 @@
 #define _WIN32_WINDOWS 0x0501
 #define _WIN32_WINNT   0x0501
 #include <SFML/Window/Win32/WindowImplWin32.hpp>
+#include <SFML/Window/WindowSettings.hpp>
 #include <SFML/Window/WindowStyle.hpp>
 #include <GL/gl.h>
 #include <SFML/Window/glext/wglext.h>
@@ -53,7 +54,6 @@ namespace priv
 unsigned int     WindowImplWin32::ourWindowCount      = 0;
 const char*      WindowImplWin32::ourClassNameA       = "SFML_Window";
 const wchar_t*   WindowImplWin32::ourClassNameW       = L"SFML_Window";
-WindowImplWin32* WindowImplWin32::ourDummyWindow      = NULL;
 WindowImplWin32* WindowImplWin32::ourFullscreenWindow = NULL;
 
 
@@ -62,9 +62,10 @@ WindowImplWin32* WindowImplWin32::ourFullscreenWindow = NULL;
 /// (creates a dummy window to provide a valid OpenGL context)
 ////////////////////////////////////////////////////////////
 WindowImplWin32::WindowImplWin32() :
-myHandle  (NULL),
-myCallback(0),
-myCursor  (NULL)
+myHandle          (NULL),
+myCallback        (0),
+myCursor          (NULL),
+myKeyRepeatEnabled(true)
 {
     // Register the window class at first call
     if (ourWindowCount == 0)
@@ -87,20 +88,21 @@ myCursor  (NULL)
 
     // Create the rendering context
     if (myHandle)
-        CreateContext(VideoMode(myWidth, myHeight, 32), false, 0);
-
-    // Save as the dummy window
-    if (!ourDummyWindow)
-        ourDummyWindow = this;
+    {
+        WindowSettings Params(0, 0, 0);
+        CreateContext(VideoMode(myWidth, myHeight, 32), Params);
+    }
 }
 
 
 ////////////////////////////////////////////////////////////
 /// Create the window implementation from an existing control
 ////////////////////////////////////////////////////////////
-WindowImplWin32::WindowImplWin32(WindowHandle Handle, int AntialiasingLevel) :
-myCallback(0),
-myCursor  (NULL)
+WindowImplWin32::WindowImplWin32(WindowHandle Handle, WindowSettings& Params) :
+myHandle          (NULL),
+myCallback        (0),
+myCursor          (NULL),
+myKeyRepeatEnabled(true)
 {
     // Save window handle
     myHandle = static_cast<HWND>(Handle);
@@ -114,10 +116,8 @@ myCursor  (NULL)
         myHeight = Rect.bottom - Rect.top;
 
         // Create the rendering context
-        VideoMode Mode = VideoMode::GetDesktopMode();
-        Mode.Width  = myWidth;
-        Mode.Height = myHeight;
-        CreateContext(Mode, false, AntialiasingLevel);
+        VideoMode Mode(myWidth, myHeight, VideoMode::GetDesktopMode().BitsPerPixel);
+        CreateContext(Mode, Params);
 
         // We change the event procedure of the control (it is important to save the old one)
         SetWindowLongPtr(myHandle, GWLP_USERDATA, reinterpret_cast<long>(this));
@@ -129,10 +129,11 @@ myCursor  (NULL)
 ////////////////////////////////////////////////////////////
 /// Create the window implementation
 ////////////////////////////////////////////////////////////
-WindowImplWin32::WindowImplWin32(VideoMode Mode, const std::string& Title, unsigned long WindowStyle, int AntialiasingLevel) :
-myHandle  (NULL),
-myCallback(0),
-myCursor  (NULL)
+WindowImplWin32::WindowImplWin32(VideoMode Mode, const std::string& Title, unsigned long WindowStyle, WindowSettings& Params) :
+myHandle          (NULL),
+myCallback        (0),
+myCursor          (NULL),
+myKeyRepeatEnabled(true)
 {
     // Register the window class at first call
     if (ourWindowCount == 0)
@@ -145,9 +146,17 @@ myCursor  (NULL)
     int Height = myHeight = Mode.Height;
 
     // Choose the window style according to the Style parameter
-    DWORD Win32Style = WS_VISIBLE | WS_CAPTION | WS_MINIMIZEBOX;
-    if (WindowStyle & Style::Resize) Win32Style |= WS_THICKFRAME | WS_MAXIMIZEBOX;
-    if (WindowStyle & Style::Close)  Win32Style |= WS_SYSMENU;
+    DWORD Win32Style = WS_VISIBLE;
+    if (WindowStyle == Style::None)
+    {
+        Win32Style |= WS_POPUP;
+    }
+    else
+    {
+        if (WindowStyle & Style::Titlebar) Win32Style |= WS_CAPTION | WS_MINIMIZEBOX;
+        if (WindowStyle & Style::Resize)   Win32Style |= WS_THICKFRAME | WS_MAXIMIZEBOX;
+        if (WindowStyle & Style::Close)    Win32Style |= WS_SYSMENU;
+    }
 
     // In windowed mode, adjust width and height so that window will have the requested client area
     bool Fullscreen = (WindowStyle & Style::Fullscreen) != 0;
@@ -172,9 +181,13 @@ myCursor  (NULL)
         myHandle = CreateWindowA(ourClassNameA, Title.c_str(), Win32Style, Left, Top, Width, Height, NULL, NULL, GetModuleHandle(NULL), this);
     }
 
+    // Switch to fullscreen if requested
+    if (Fullscreen)
+        SwitchToFullscreen(Mode);
+
     // Create the rendering context
     if (myHandle)
-        CreateContext(Mode, Fullscreen, AntialiasingLevel);
+        CreateContext(Mode, Params);
 
     // Increment window count
     ourWindowCount++;
@@ -216,6 +229,24 @@ WindowImplWin32::~WindowImplWin32()
 }
 
 
+////////////////////////////////////////////////////////////
+/// /see sfWindowImpl::ProcessEvents
+////////////////////////////////////////////////////////////
+void WindowImplWin32::ProcessEvents()
+{
+    // We update the window only if we own it
+    if (!myCallback)
+    {
+        MSG Message;
+        while (PeekMessage(&Message, myHandle, 0, 0, PM_REMOVE))
+        {
+            TranslateMessage(&Message);
+            DispatchMessage(&Message);
+        }
+    }
+}
+
+
 ////////////////////////////////////////////////////////////
 /// /see sfWindowImpl::Display
 ////////////////////////////////////////////////////////////
@@ -227,30 +258,19 @@ void WindowImplWin32::Display()
 
 
 ////////////////////////////////////////////////////////////
-/// /see sfWindowImpl::ProcessEvents
+/// /see sfWindowImpl::MakeActive
 ////////////////////////////////////////////////////////////
-void WindowImplWin32::ProcessEvents()
+void WindowImplWin32::MakeActive(bool Active) const
 {
-    // We update the window only if we own it
-    if (!myCallback)
+    if (Active)
     {
-        MSG Message;
-        while (PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
-        {
-            TranslateMessage(&Message);
-            DispatchMessage(&Message);
-        }
+        if (myDeviceContext && myGLContext)
+            wglMakeCurrent(myDeviceContext, myGLContext);
+    }
+    else
+    {
+        wglMakeCurrent(NULL, NULL);
     }
-}
-
-
-////////////////////////////////////////////////////////////
-/// /see sfWindowImpl::MakeCurrent
-////////////////////////////////////////////////////////////
-void WindowImplWin32::MakeCurrent() const
-{
-    if (myDeviceContext && myGLContext)
-        wglMakeCurrent(myDeviceContext, myGLContext);
 }
 
 
@@ -308,6 +328,15 @@ void WindowImplWin32::Show(bool State)
 }
 
 
+////////////////////////////////////////////////////////////
+/// /see sfWindowImpl::EnableKeyRepeat
+////////////////////////////////////////////////////////////
+void WindowImplWin32::EnableKeyRepeat(bool Enabled)
+{
+    myKeyRepeatEnabled = Enabled;
+}
+
+
 ////////////////////////////////////////////////////////////
 /// Register the window class
 ////////////////////////////////////////////////////////////
@@ -347,43 +376,46 @@ void WindowImplWin32::RegisterWindowClass()
 
 
 ////////////////////////////////////////////////////////////
-/// Create the OpenGL rendering context
+/// Switch to fullscreen mode
 ////////////////////////////////////////////////////////////
-void WindowImplWin32::CreateContext(VideoMode Mode, bool Fullscreen, int AntialiasingLevel)
+void WindowImplWin32::SwitchToFullscreen(const VideoMode& Mode)
 {
-    // Fullscreen mode
-    if (Fullscreen)
+    DEVMODE DevMode;
+    DevMode.dmSize       = sizeof(DEVMODE);
+    DevMode.dmPelsWidth  = Mode.Width;
+    DevMode.dmPelsHeight = Mode.Height;
+    DevMode.dmBitsPerPel = Mode.BitsPerPixel;
+    DevMode.dmFields     = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
+
+    // Apply fullscreen mode
+    if (ChangeDisplaySettings(&DevMode, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
     {
-        DEVMODE DevMode;
-        DevMode.dmSize       = sizeof(DEVMODE);
-        DevMode.dmPelsWidth  = Mode.Width;
-        DevMode.dmPelsHeight = Mode.Height;
-        DevMode.dmBitsPerPel = Mode.BitsPerPixel;
-        DevMode.dmFields     = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
-
-        // Apply fullscreen mode
-        if (ChangeDisplaySettings(&DevMode, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
-        {
-            std::cerr << "Failed to change display mode for fullscreen -- cannot create OpenGL context" << std::endl;
-            return;
-        }
-
-        // Change window style (no border, no titlebar, ...)
-        SetWindowLong(myHandle, GWL_STYLE,   WS_POPUP);
-        SetWindowLong(myHandle, GWL_EXSTYLE, WS_EX_APPWINDOW);
-
-        // And resize it so that it fits the entire screen
-        SetWindowPos(myHandle, HWND_TOP, 0, 0, Mode.Width, Mode.Height, SWP_FRAMECHANGED);
-        ShowWindow(myHandle, SW_SHOW);
-
-        // Set "this" as the current fullscreen window
-        ourFullscreenWindow = this;
+        std::cerr << "Failed to change display mode for fullscreen" << std::endl;
+        return;
     }
 
+    // Change window style (no border, no titlebar, ...)
+    SetWindowLong(myHandle, GWL_STYLE,   WS_POPUP);
+    SetWindowLong(myHandle, GWL_EXSTYLE, WS_EX_APPWINDOW);
+
+    // And resize it so that it fits the entire screen
+    SetWindowPos(myHandle, HWND_TOP, 0, 0, Mode.Width, Mode.Height, SWP_FRAMECHANGED);
+    ShowWindow(myHandle, SW_SHOW);
+
+    // Set "this" as the current fullscreen window
+    ourFullscreenWindow = this;
+
     // SetPixelFormat can fail (really ?) if window style doesn't contain these flags
     long Style = GetWindowLong(myHandle, GWL_STYLE);
     SetWindowLong(myHandle, GWL_STYLE, Style | WS_CLIPCHILDREN | WS_CLIPSIBLINGS);
+}
 
+
+////////////////////////////////////////////////////////////
+/// Construct the context from graphics settings
+////////////////////////////////////////////////////////////
+void WindowImplWin32::CreateContext(const VideoMode& Mode, WindowSettings& Params)
+{
     // Get the device context attached to the window
     myDeviceContext = GetDC(myHandle);
     if (myDeviceContext == NULL)
@@ -393,8 +425,8 @@ void WindowImplWin32::CreateContext(VideoMode Mode, bool Fullscreen, int Antiali
     }
 
     // Let's find a suitable pixel format -- first try with antialiasing
-    int PixelFormat = 0;
-    if (AntialiasingLevel > 0)
+    int BestFormat = 0;
+    if (Params.AntialiasingLevel > 0)
     {
         // Get the wglChoosePixelFormatARB function (it is an extension)
         PFNWGLCHOOSEPIXELFORMATARBPROC wglChoosePixelFormatARB = reinterpret_cast<PFNWGLCHOOSEPIXELFORMATARBPROC>(wglGetProcAddress("wglChoosePixelFormatARB"));
@@ -402,30 +434,29 @@ void WindowImplWin32::CreateContext(VideoMode Mode, bool Fullscreen, int Antiali
         // Define the basic attributes we want for our window
         int IntAttributes[] =
         {
-		    WGL_SAMPLES_ARB,        AntialiasingLevel,
-		    WGL_SAMPLE_BUFFERS_ARB, GL_TRUE,
             WGL_DRAW_TO_WINDOW_ARB, GL_TRUE,
 		    WGL_SUPPORT_OPENGL_ARB, GL_TRUE,
 		    WGL_ACCELERATION_ARB,   WGL_FULL_ACCELERATION_ARB,
-		    WGL_COLOR_BITS_ARB,     Mode.BitsPerPixel,
 		    WGL_DOUBLE_BUFFER_ARB,  GL_TRUE,
+            WGL_SAMPLE_BUFFERS_ARB, (Params.AntialiasingLevel ? GL_TRUE : GL_FALSE),
+		    WGL_SAMPLES_ARB,        Params.AntialiasingLevel,
 		    0,                      0
         };
 
         // Let's check how many formats are supporting our requirements
-        int   Formats[64];
+        int   Formats[128];
 	    UINT  NbFormats;
 	    float FloatAttributes[] = {0, 0};
 	    bool  IsValid = wglChoosePixelFormatARB(myDeviceContext, IntAttributes, FloatAttributes, sizeof(Formats) / sizeof(*Formats), Formats, &NbFormats) != 0;
         if (!IsValid || (NbFormats == 0))
         {
-            if (AntialiasingLevel > 2)
+            if (Params.AntialiasingLevel > 2)
             {
                 // No format matching our needs : reduce the multisampling level
                 std::cerr << "Failed to find a pixel format supporting "
-                          << AntialiasingLevel << " antialiasing levels ; trying with 2 levels" << std::endl;
+                          << Params.AntialiasingLevel << " antialiasing levels ; trying with 2 levels" << std::endl;
 
-                IntAttributes[1] = 2;
+                Params.AntialiasingLevel = IntAttributes[1] = 2;
 	            IsValid = wglChoosePixelFormatARB(myDeviceContext, IntAttributes, FloatAttributes, sizeof(Formats) / sizeof(*Formats), Formats, &NbFormats) != 0;
             }
 
@@ -433,19 +464,15 @@ void WindowImplWin32::CreateContext(VideoMode Mode, bool Fullscreen, int Antiali
             {
                 // Cannot find any pixel format supporting multisampling ; disabling antialiasing
                 std::cerr << "Failed to find a pixel format supporting antialiasing ; antialiasing will be disabled" << std::endl;
-                AntialiasingLevel = 0;
+                Params.AntialiasingLevel = 0;
             }
         }
 
         // Get the best format among the returned ones
         if (IsValid && (NbFormats > 0))
         {
-            PixelFormat = Formats[0];
-            PIXELFORMATDESCRIPTOR BestAttribs;
-            BestAttribs.nSize    = sizeof(PIXELFORMATDESCRIPTOR);
-            BestAttribs.nVersion = 1;
-            DescribePixelFormat(myDeviceContext, PixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &BestAttribs);
-            for (UINT i = 1; i < NbFormats; ++i)
+            int BestScore = 0xFFFF;
+            for (UINT i = 0; i < NbFormats; ++i)
             {
                 // Get the current format's attributes
                 PIXELFORMATDESCRIPTOR Attribs;
@@ -453,18 +480,22 @@ void WindowImplWin32::CreateContext(VideoMode Mode, bool Fullscreen, int Antiali
                 Attribs.nVersion = 1;
                 DescribePixelFormat(myDeviceContext, Formats[i], sizeof(PIXELFORMATDESCRIPTOR), &Attribs);
 
-                // Compare them with the current best format
-                if ((BestAttribs.cStencilBits < Attribs.cStencilBits) || (BestAttribs.cDepthBits < Attribs.cDepthBits))
+                // Evaluate the current configuration
+                int Color = Attribs.cRedBits + Attribs.cGreenBits + Attribs.cBlueBits + Attribs.cAlphaBits;
+                int Score = EvaluateConfig(Mode, Params, Color, Attribs.cDepthBits, Attribs.cStencilBits, Params.AntialiasingLevel);
+
+                // Keep it if it's better than the current best
+                if (Score < BestScore)
                 {
-                    PixelFormat = Formats[i];
-                    BestAttribs = Attribs;
+                    BestScore  = Score;
+                    BestFormat = Formats[i];
                 }
             }
         }
     }
 
     // Find a pixel format with no antialiasing, if not needed or not supported
-    if (AntialiasingLevel == 0)
+    if (BestFormat == 0)
     {
         // Setup a pixel format descriptor from the rendering settings
         PIXELFORMATDESCRIPTOR PixelDescriptor;
@@ -475,12 +506,12 @@ void WindowImplWin32::CreateContext(VideoMode Mode, bool Fullscreen, int Antiali
         PixelDescriptor.dwFlags      = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
         PixelDescriptor.iPixelType   = PFD_TYPE_RGBA;
         PixelDescriptor.cColorBits   = static_cast<BYTE>(Mode.BitsPerPixel);
-        PixelDescriptor.cDepthBits   = 32; // It's only a hint
-        PixelDescriptor.cStencilBits = 8;  // It's only a hint
+        PixelDescriptor.cDepthBits   = static_cast<BYTE>(Params.DepthBits);
+        PixelDescriptor.cStencilBits = static_cast<BYTE>(Params.StencilBits);
 
         // Get the pixel format that best matches our requirements
-        PixelFormat = ChoosePixelFormat(myDeviceContext, &PixelDescriptor);
-        if (PixelFormat == 0)
+        BestFormat = ChoosePixelFormat(myDeviceContext, &PixelDescriptor);
+        if (BestFormat == 0)
         {
             std::cerr << "Failed to find a suitable pixel format for device context -- cannot create OpenGL context" << std::endl;
             return;
@@ -488,15 +519,15 @@ void WindowImplWin32::CreateContext(VideoMode Mode, bool Fullscreen, int Antiali
     }
 
     // Extract the depth and stencil bits from the chosen format
-    PIXELFORMATDESCRIPTOR ActualDescriptor;
-    ActualDescriptor.nSize    = sizeof(PIXELFORMATDESCRIPTOR);
-    ActualDescriptor.nVersion = 1;
-    DescribePixelFormat(myDeviceContext, PixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &ActualDescriptor);
-    myDepthBits   = ActualDescriptor.cDepthBits;
-    myStencilBits = ActualDescriptor.cStencilBits;
+    PIXELFORMATDESCRIPTOR ActualFormat;
+    ActualFormat.nSize    = sizeof(PIXELFORMATDESCRIPTOR);
+    ActualFormat.nVersion = 1;
+    DescribePixelFormat(myDeviceContext, BestFormat, sizeof(PIXELFORMATDESCRIPTOR), &ActualFormat);
+    Params.DepthBits   = ActualFormat.cDepthBits;
+    Params.StencilBits = ActualFormat.cStencilBits;
 
     // Set the chosen pixel format
-    if (!SetPixelFormat(myDeviceContext, PixelFormat, &ActualDescriptor))
+    if (!SetPixelFormat(myDeviceContext, BestFormat, &ActualFormat))
     {
         std::cerr << "Failed to set pixel format for device context -- cannot create OpenGL context" << std::endl;
         return;
@@ -514,11 +545,11 @@ void WindowImplWin32::CreateContext(VideoMode Mode, bool Fullscreen, int Antiali
     if (wglGetCurrentContext())
         wglShareLists(wglGetCurrentContext(), myGLContext);
 
-    // Set our context as the current OpenGL context for rendering
-    SetCurrent();
+    // Activate the context
+    MakeActive(true);
 
     // Enable multisampling
-    if (AntialiasingLevel > 0)
+    if (Params.AntialiasingLevel > 0)
         glEnable(GL_MULTISAMPLE_ARB);
 }
 
@@ -541,24 +572,9 @@ void WindowImplWin32::Cleanup()
     // Destroy the OpenGL context
     if (myGLContext)
     {
-        // If this is not the dummy window, we must set it as the valid rendering context to avoid a crash with next OpenGL command
-        if (this != ourDummyWindow)
-        {
-            if (ourDummyWindow)
-                ourDummyWindow->SetCurrent();
-        }
-        else
-        {
-            ourDummyWindow = NULL;
-            wglMakeCurrent(NULL, NULL);
-        }
-
-        // Destroy the context
         wglDeleteContext(myGLContext);
         myGLContext = NULL;
     }
-
-    // Release the device context
     if (myDeviceContext)
     {
         ReleaseDC(myHandle, myDeviceContext);
@@ -654,13 +670,16 @@ void WindowImplWin32::ProcessEvent(UINT Message, WPARAM WParam, LPARAM LParam)
         case WM_KEYDOWN :
         case WM_SYSKEYDOWN :
         {
-            Event Evt;
-            Evt.Type        = Event::KeyPressed;
-            Evt.Key.Code    = VirtualKeyCodeToSF(WParam);
-            Evt.Key.Alt     = HIWORD(GetAsyncKeyState(VK_MENU))    != 0;
-            Evt.Key.Control = HIWORD(GetAsyncKeyState(VK_CONTROL)) != 0;
-            Evt.Key.Shift   = HIWORD(GetAsyncKeyState(VK_SHIFT))   != 0;
-            SendEvent(Evt);
+            if (myKeyRepeatEnabled || ((LParam & (1 << 30)) == 0))
+            {
+                Event Evt;
+                Evt.Type        = Event::KeyPressed;
+                Evt.Key.Code    = (WParam == VK_SHIFT) ? GetShiftState(true) : VirtualKeyCodeToSF(WParam, LParam);
+                Evt.Key.Alt     = HIWORD(GetAsyncKeyState(VK_MENU))    != 0;
+                Evt.Key.Control = HIWORD(GetAsyncKeyState(VK_CONTROL)) != 0;
+                Evt.Key.Shift   = HIWORD(GetAsyncKeyState(VK_SHIFT))   != 0;
+                SendEvent(Evt);
+            }
             break;
         }
 
@@ -670,7 +689,7 @@ void WindowImplWin32::ProcessEvent(UINT Message, WPARAM WParam, LPARAM LParam)
         {
             Event Evt;
             Evt.Type        = Event::KeyReleased;
-            Evt.Key.Code    = VirtualKeyCodeToSF(WParam);
+            Evt.Key.Code    = (WParam == VK_SHIFT) ? GetShiftState(false) : VirtualKeyCodeToSF(WParam, LParam);
             Evt.Key.Alt     = HIWORD(GetAsyncKeyState(VK_MENU))    != 0;
             Evt.Key.Control = HIWORD(GetAsyncKeyState(VK_CONTROL)) != 0;
             Evt.Key.Shift   = HIWORD(GetAsyncKeyState(VK_SHIFT))   != 0;
@@ -782,97 +801,142 @@ void WindowImplWin32::ProcessEvent(UINT Message, WPARAM WParam, LPARAM LParam)
 }
 
 
+////////////////////////////////////////////////////////////
+/// Check the state of the shift keys on a key event,
+/// and return the corresponding SF key code
+////////////////////////////////////////////////////////////
+Key::Code WindowImplWin32::GetShiftState(bool KeyDown)
+{
+    static bool LShiftPrevDown = false;
+    static bool RShiftPrevDown = false;
+
+    bool LShiftDown = (HIWORD(GetAsyncKeyState(VK_LSHIFT)) != 0);
+    bool RShiftDown = (HIWORD(GetAsyncKeyState(VK_RSHIFT)) != 0);
+
+    Key::Code Code = Key::Code(0);
+    if (KeyDown)
+    {
+        if      (!LShiftPrevDown && LShiftDown) Code = Key::LShift;
+        else if (!RShiftPrevDown && RShiftDown) Code = Key::RShift;
+    }
+    else
+    {
+        if      (LShiftPrevDown && !LShiftDown) Code = Key::LShift;
+        else if (RShiftPrevDown && !RShiftDown) Code = Key::RShift;
+    }
+
+    LShiftPrevDown = LShiftDown;
+    RShiftPrevDown = RShiftDown;
+
+    return Code;
+}
+
+
 ////////////////////////////////////////////////////////////
 /// Convert a Win32 virtual key code to a SFML key code
 ////////////////////////////////////////////////////////////
-Key::Code WindowImplWin32::VirtualKeyCodeToSF(WPARAM VirtualKey)
+Key::Code WindowImplWin32::VirtualKeyCodeToSF(WPARAM VirtualKey, LPARAM Flags)
 {
     switch (VirtualKey)
     {
-        case VK_ESCAPE :   return Key::Escape;
-        case VK_MENU :     return Key::Alt;
-        case VK_CONTROL :  return Key::Control;
-        case VK_SHIFT :    return Key::Shift;
-        case VK_SPACE :    return Key::Space;
-        case VK_RETURN :   return Key::Return;
-        case VK_BACK :     return Key::Back;
-        case VK_TAB :      return Key::Tab;
-        case VK_PRIOR :    return Key::PageUp;
-        case VK_NEXT :     return Key::PageDown;
-        case VK_END :      return Key::End;
-        case VK_HOME :     return Key::Home;
-        case VK_INSERT :   return Key::Insert;
-        case VK_DELETE :   return Key::Delete;
-        case VK_ADD :      return Key::Add;
-        case VK_SUBTRACT : return Key::Subtract;
-        case VK_MULTIPLY : return Key::Multiply;
-        case VK_DIVIDE :   return Key::Divide;
-        case VK_PAUSE :    return Key::Pause;
-        case VK_F1 :       return Key::F1;
-        case VK_F2 :       return Key::F2;
-        case VK_F3 :       return Key::F3;
-        case VK_F4 :       return Key::F4;
-        case VK_F5 :       return Key::F5;
-        case VK_F6 :       return Key::F6;
-        case VK_F7 :       return Key::F7;
-        case VK_F8 :       return Key::F8;
-        case VK_F9 :       return Key::F9;
-        case VK_F10 :      return Key::F10;
-        case VK_F11 :      return Key::F11;
-        case VK_F12 :      return Key::F12;
-        case VK_F13 :      return Key::F13;
-        case VK_F14 :      return Key::F14;
-        case VK_F15 :      return Key::F15;
-        case VK_LEFT :     return Key::Left;
-        case VK_RIGHT :    return Key::Right;
-        case VK_UP :       return Key::Up;
-        case VK_DOWN :     return Key::Down;
-        case VK_NUMPAD0 :  return Key::Numpad0;
-        case VK_NUMPAD1 :  return Key::Numpad1;
-        case VK_NUMPAD2 :  return Key::Numpad2;
-        case VK_NUMPAD3 :  return Key::Numpad3;
-        case VK_NUMPAD4 :  return Key::Numpad4;
-        case VK_NUMPAD5 :  return Key::Numpad5;
-        case VK_NUMPAD6 :  return Key::Numpad6;
-        case VK_NUMPAD7 :  return Key::Numpad7;
-        case VK_NUMPAD8 :  return Key::Numpad8;
-        case VK_NUMPAD9 :  return Key::Numpad9;
-        case 'A' :         return Key::A;
-        case 'Z' :         return Key::Z;
-        case 'E' :         return Key::E;
-        case 'R' :         return Key::R;
-        case 'T' :         return Key::T;
-        case 'Y' :         return Key::Y;
-        case 'U' :         return Key::U;
-        case 'I' :         return Key::I;
-        case 'O' :         return Key::O;
-        case 'P' :         return Key::P;
-        case 'Q' :         return Key::Q;
-        case 'S' :         return Key::S;
-        case 'D' :         return Key::D;
-        case 'F' :         return Key::F;
-        case 'G' :         return Key::G;
-        case 'H' :         return Key::H;
-        case 'J' :         return Key::J;
-        case 'K' :         return Key::K;
-        case 'L' :         return Key::L;
-        case 'M' :         return Key::M;
-        case 'W' :         return Key::W;
-        case 'X' :         return Key::X;
-        case 'C' :         return Key::C;
-        case 'V' :         return Key::V;
-        case 'B' :         return Key::B;
-        case 'N' :         return Key::N;
-        case '0' :         return Key::Num0;
-        case '1' :         return Key::Num1;
-        case '2' :         return Key::Num2;
-        case '3' :         return Key::Num3;
-        case '4' :         return Key::Num4;
-        case '5' :         return Key::Num5;
-        case '6' :         return Key::Num6;
-        case '7' :         return Key::Num7;
-        case '8' :         return Key::Num8;
-        case '9' :         return Key::Num9;
+        // VK_SHIFT is handled by the GetShiftState function
+        case VK_MENU :       return (Flags & (1 << 24)) ? Key::RAlt     : Key::LAlt;
+        case VK_CONTROL :    return (Flags & (1 << 24)) ? Key::RControl : Key::LControl;
+        case VK_LWIN :       return Key::LSystem;
+        case VK_RWIN :       return Key::RSystem;
+        case VK_APPS :       return Key::Menu;
+        case VK_OEM_1 :      return Key::SemiColon;
+        case VK_OEM_2 :      return Key::Slash;
+        case VK_OEM_PLUS :   return Key::Equal;
+        case VK_OEM_MINUS :  return Key::Dash;
+        case VK_OEM_4 :      return Key::LBracket;
+        case VK_OEM_6 :      return Key::RBracket;
+        case VK_OEM_COMMA :  return Key::Comma;
+        case VK_OEM_PERIOD : return Key::Period;
+        case VK_OEM_7 :      return Key::Quote;
+        case VK_OEM_5 :      return Key::BackSlash;
+        case VK_OEM_3 :      return Key::Tilde;
+        case VK_ESCAPE :     return Key::Escape;
+        case VK_SPACE :      return Key::Space;
+        case VK_RETURN :     return Key::Return;
+        case VK_BACK :       return Key::Back;
+        case VK_TAB :        return Key::Tab;
+        case VK_PRIOR :      return Key::PageUp;
+        case VK_NEXT :       return Key::PageDown;
+        case VK_END :        return Key::End;
+        case VK_HOME :       return Key::Home;
+        case VK_INSERT :     return Key::Insert;
+        case VK_DELETE :     return Key::Delete;
+        case VK_ADD :        return Key::Add;
+        case VK_SUBTRACT :   return Key::Subtract;
+        case VK_MULTIPLY :   return Key::Multiply;
+        case VK_DIVIDE :     return Key::Divide;
+        case VK_PAUSE :      return Key::Pause;
+        case VK_F1 :         return Key::F1;
+        case VK_F2 :         return Key::F2;
+        case VK_F3 :         return Key::F3;
+        case VK_F4 :         return Key::F4;
+        case VK_F5 :         return Key::F5;
+        case VK_F6 :         return Key::F6;
+        case VK_F7 :         return Key::F7;
+        case VK_F8 :         return Key::F8;
+        case VK_F9 :         return Key::F9;
+        case VK_F10 :        return Key::F10;
+        case VK_F11 :        return Key::F11;
+        case VK_F12 :        return Key::F12;
+        case VK_F13 :        return Key::F13;
+        case VK_F14 :        return Key::F14;
+        case VK_F15 :        return Key::F15;
+        case VK_LEFT :       return Key::Left;
+        case VK_RIGHT :      return Key::Right;
+        case VK_UP :         return Key::Up;
+        case VK_DOWN :       return Key::Down;
+        case VK_NUMPAD0 :    return Key::Numpad0;
+        case VK_NUMPAD1 :    return Key::Numpad1;
+        case VK_NUMPAD2 :    return Key::Numpad2;
+        case VK_NUMPAD3 :    return Key::Numpad3;
+        case VK_NUMPAD4 :    return Key::Numpad4;
+        case VK_NUMPAD5 :    return Key::Numpad5;
+        case VK_NUMPAD6 :    return Key::Numpad6;
+        case VK_NUMPAD7 :    return Key::Numpad7;
+        case VK_NUMPAD8 :    return Key::Numpad8;
+        case VK_NUMPAD9 :    return Key::Numpad9;
+        case 'A' :           return Key::A;
+        case 'Z' :           return Key::Z;
+        case 'E' :           return Key::E;
+        case 'R' :           return Key::R;
+        case 'T' :           return Key::T;
+        case 'Y' :           return Key::Y;
+        case 'U' :           return Key::U;
+        case 'I' :           return Key::I;
+        case 'O' :           return Key::O;
+        case 'P' :           return Key::P;
+        case 'Q' :           return Key::Q;
+        case 'S' :           return Key::S;
+        case 'D' :           return Key::D;
+        case 'F' :           return Key::F;
+        case 'G' :           return Key::G;
+        case 'H' :           return Key::H;
+        case 'J' :           return Key::J;
+        case 'K' :           return Key::K;
+        case 'L' :           return Key::L;
+        case 'M' :           return Key::M;
+        case 'W' :           return Key::W;
+        case 'X' :           return Key::X;
+        case 'C' :           return Key::C;
+        case 'V' :           return Key::V;
+        case 'B' :           return Key::B;
+        case 'N' :           return Key::N;
+        case '0' :           return Key::Num0;
+        case '1' :           return Key::Num1;
+        case '2' :           return Key::Num2;
+        case '3' :           return Key::Num3;
+        case '4' :           return Key::Num4;
+        case '5' :           return Key::Num5;
+        case '6' :           return Key::Num6;
+        case '7' :           return Key::Num7;
+        case '8' :           return Key::Num8;
+        case '9' :           return Key::Num9;
     }
 
     return Key::Code(0);
@@ -928,6 +992,10 @@ LRESULT CALLBACK WindowImplWin32::GlobalOnEvent(HWND Handle, UINT Message, WPARA
             return CallWindowProc(reinterpret_cast<WNDPROC>(Window->myCallback), Handle, Message, WParam, LParam);
     }
 
+    // We don't forward the WM_CLOSE message to prevent the OS to automatically destroy the window
+    if (Message == WM_CLOSE)
+        return 0;
+
     static const bool HasUnicode = HasUnicodeSupport();
     return HasUnicode ? DefWindowProcW(Handle, Message, WParam, LParam) :
                         DefWindowProcA(Handle, Message, WParam, LParam);
